from __future__ import annotations

import logging
from typing import List, Optional

from src import errors
from src import utils


class ErrorMsg(errors.ErrorMsgBase):
    MOOD_GROUP_NOT_FOUND = "Expected to find {} mood group, but the key is missing from the dictionary."
    STANDARD_MOODS_USED = "Problem parsing custom moods file. Standard mood set - i.e. {} - will be used."
    SKIPPED_INVALID_MOOD = "Skipping \'{}\' as it is not a valid mood."


class EmptyValue(utils.CustomException):
    """Attribute cannot be set to empty."""


class MoodNotFoundError(utils.CustomException):
    """The mood could not be found neither in the standard mood set nor the custom one."""


# noinspection SpellCheckingInspection
class Moodverse:
    """
    Moodverse is the single source of truth regarding moods. It always knows the bare-minimum set of moods.
    Its knowledge can be expanded with custom mood sets.
    """
    def __init__(self, moods_to_process: dict[str, List[str]] = None):
        """
        If you want to expand the standard mood set with your custom one, make sure to pass a valid mood set file.
        Even one error in the dict causes Moodverse to reject it, and it will then proceed to use just the standard set.
        :param moods_to_process: Lists of moods grouped into five dictionary keys: rad, great, neutral, bad & awful
        """
        self.__logger = logging.getLogger(self.__class__.__name__)
        # DEFAULT PART OF INIT
        # --------------------
        # Build a minimal-viable mood set with these five mood groups
        # .
        # ├── known moods of 'rad' group
        # │   └── rad
        # ├── known moods of 'great' group
        # │   └── great
        # ├── known moods of 'neutral' group
        # │   └── neutral
        # ├── known moods of 'bad' group
        # │   └── bad
        # └── known moods of 'awful' group
        #     └── awful

        self.__mood_set: dict[str, MoodGroup] = {}
        self.__has_custom_moods = False

        for group_name in ["rad", "good", "neutral", "bad", "awful"]:
            # Add the new group to the dict
            self.__mood_set[group_name] = MoodGroup(group_name)
            # Ask it to create its main mood (e.g. 'rad' for 'rad'-group)
            self.__mood_set[group_name].create_mood()

        # We can stop here and be content with our "default" / "standard" mood-set if user did not pass a custom one
        # CUSTOM PART OF INIT
        # --------------------
        if moods_to_process is not None:
            try:
                self.__expand_moodset_with_customs(moods_to_process)
            except MoodNotFoundError:
                msg = ErrorMsg.print(ErrorMsg.STANDARD_MOODS_USED, ', '.join(self.__mood_set.keys()))
                self.__logger.warning(msg)
            else:
                self.__has_custom_moods = True

    def __expand_moodset_with_customs(self, moods_to_process: dict[str, List[str]]) -> None:
        """
        Expands the knowledge of this specific :class:`Moodverse` with new custom moods passed by user.
        :param moods_to_process: Lists of str moods grouped into five dictionary keys: rad, great, neutral, bad & awful
        :raises MoodNotFoundError: if at least one mood was found to be invalid, no new moods will be saved
        """
        # Then expand the minimal-viable mood set from Moodverse __init__
        # .
        # ├── known moods of 'rad' group
        # │   └── rad
        # │   └── (add more...)
        # ├── known moods of 'great' group
        # │   └── great
        # │   └── (add more...)
        # ├── known moods of 'neutral' group
        # │   └── neutral
        # │   └── (add more...)
        # ├── known moods of 'bad' group
        # │   └── bad
        # │   └── (add more...)
        # └── known moods of 'awful' group
        #     └── awful
        #     └── (add more...)

        # Use keys from standard_mood_set to navigate through the passed dictionary and add their unique moods over
        # The side effect is that keys in the passed dictionary which do not appear in standard mood set are skipped

        # e.g. I'm expecting a "rad" group to be in the dict
        for expected_mood_group in self.__mood_set.keys():
            expected_mood_group: str
            try:
                # Leap of faith - there must be a "rad" group, otherwise there's no point in continuing
                mood_group_to_transfer = moods_to_process[expected_mood_group]
            except KeyError as err:
                msg = ErrorMsg.print(ErrorMsg.MOOD_GROUP_NOT_FOUND, expected_mood_group)
                self.__logger.warning(msg)
                raise MoodNotFoundError(msg) from err
            # Go through each mood in this mood group - e.g. rad - and transfer them in the form of Mood objects
            for mood in mood_group_to_transfer:
                self.__mood_set[expected_mood_group].create_mood(mood)

    def get_mood(self, value_to_check: str) -> Optional['Mood']:
        """
        Checks if the mood exists in the currently used mood set. None if it does not.
        :param value_to_check: string with the name of the mood to find
        :return: reference to the :class:`Mood` object found in the currently used mood set or None
        """
        return next((mood_group.known_moods[value_to_check] for mood_group in self.__mood_set.values() if
                     value_to_check in mood_group.known_moods), None)

    def __getitem__(self, item: str):
        return self.__mood_set[item]

    @property
    def known_mood_groups(self):
        return self.__mood_set

    @property
    def has_custom_moods(self):
        return self.__has_custom_moods


class AbstractMood:
    """
    Provides shared methods for :class:`MoodGroup` and :class:`Mood`.
    """
    def __init__(self, value):
        if not value or isinstance(value, str) is False:
            raise ValueError
        self.__name = value

    @property
    def name(self) -> str:
        return self.__name

    def __str__(self) -> str:
        return self.__name

    def __eq__(self, other) -> bool:
        if isinstance(other, str):
            return str(self) == other
        return super().__eq__(other)  # of object object


class MoodGroup(AbstractMood):
    """
    Mood group is an abstract way of thinking about moods, based on categorising them into different groups (or levels).

    **For example**:
    Feeling 'awesome' and 'amazing' all belong to the same group.
    Therefore, if we categorise moods to different categories, we create mood groups.
    They could, for example, range from 5 (best) to 1 (worst).

    Daylio uses 'rad', 'great', 'neutral', 'bad' & 'awful'.
    """
    def __init__(self, name_of_the_mood_group: str):
        """
        Create a :class:`MoodGroup` object with the specified name.
        :param name_of_the_mood_group: Name of the mood group you want to create.
        """
        self.__logger = logging.getLogger(self.__class__.__name__)
        self.__known_moods: dict[str, Mood] = {}

        super().__init__(name_of_the_mood_group)

    def create_mood(self, name: str = None) -> 'Mood':
        """
        Create the specified mood and append its reference to the known moods in this group.
        :param name: Name of the mood. If none provided, use the mood group name as its name (e.g. rad group -> rad).
        :return: reference to the newly created :class:`Mood` object
        """
        # e.g. if no argument is passed, MoodGroup "rad" will create a Mood "rad"
        # it's just a shorthand so that you don't have to write MoodGroup("rad").create_mood("rad")
        # └── known moods of 'rad' group
        #     └── rad
        final_name = self.name if name is None else name
        try:
            ref = Mood(final_name)
        except (EmptyValue, ValueError):
            self.__logger.warning(ErrorMsg.print(ErrorMsg.SKIPPED_INVALID_MOOD, final_name))
        # all went ok
        else:
            self.__known_moods[final_name] = ref
            return ref

    # TODO: possibly could do funky stuff with multiple inheritance - this method could come from Moodverse
    @property
    def known_moods(self) -> dict[str, 'Mood']:
        return self.__known_moods

    # TODO: possibly could do funky stuff with multiple inheritance - this method could come from Moodverse
    # I cannot type hint that this method returns a Mood object because of evaluation of annotations problem
    # It is discussed here: https://peps.python.org/pep-0563/
    # I could resolve it with ``from __future__ import annotations``, however it requires more recent Python versions
    # Right now I go around the problem by providing the class name in apostrophes
    def __getitem__(self, item: str) -> 'Mood':
        """
        Access the mood in this mood group.
        :raises KeyError: there is no such mood in this mood group.
        :param item: Which mood would you like to access?
        :return: reference to the mood object
        """
        if item in self.__known_moods:
            return self.__known_moods[item]
        raise KeyError

    def __eq__(self, other: List[str]) -> bool:
        """
        Does this mood group contain exactly the same moods as in the passed list.
        If other types of objects are passed rather than List[str], call the higher-ups.
        :param other: list of moods in string format to compare the mood group to
        :return: bool
        """
        # Used to compare instance of this class to ["mood", "mood", "mood"] and check if they contain the same moods
        if isinstance(other, list) and all(isinstance(item, str) for item in other):
            # I'm not sure why, but set() instead of pure array makes sure that the order is irrelevant
            # therefore ["nice", "good"] == ["good", "nice"] is Truthy, as expected
            return set([str(obj) for obj in self.known_moods]) == set(other)
        # Call the superclass' __eq__ for any other comparison
        return super().__eq__(other)


class Mood(AbstractMood):
    """
    A specific mood that belongs to a specific mood group.

    *For example*:

    - rad is a mood. It belongs to rad group.
    - awesome is a mood. It also belongs to the rad group.
    - hungry is a mood. It belongs either to neutral, bad or awful mood groups, depending on user preferences I guess"""
    def __init__(self, value: str):
        super().__init__(value)



"""
Librarian is a builder/singleton type class. It creates and initialises other builder objects - e.g. DatedEntriesGroup.
It sets up the process, parses the CSV file and passes extracted values to DatedEntriesGroup.
Imagine Librarian is an actual person, reading the contents of the file out-loud to a scribe (DatedEntriesGroup).
Each date is a different scribe, but they all listen to the same Librarian.
Librarian knows their identity and can call upon them when needed to recite their contents back to the Librarian.

Here's a quick breakdown of what is the specialisation of this file in the journaling process:

``└── ALL NOTES``
    ``└── all notes written on a particular date``
        ``└── a particular note``
"""
from __future__ import annotations

import csv
import json
import os
import logging
from typing import IO

from src import dated_entries_group
from src import errors
from src import utils
from src.config import options
from src.dated_entries_group import DatedEntriesGroup
from src.entry.mood import Moodverse

# Adding Librarian-specific options in global_settings
librarian_settings = options.arg_console.add_argument_group(
    "Librarian",
    "Handles main options"
)
# 1. Filepath is absolutely crucial to even start processing
librarian_settings.add_argument(
    "filepath",
    type=str,
    help="Specify path to the .CSV file"
)
# 2. Destination is not needed if user is only processing, but no destination makes it impossible to output that data.
librarian_settings.add_argument(
    "destination",
    type=str,
    help="Path to folder to output finished files into."
)
# TODO: Force-argument does nothing yet.
librarian_settings.add_argument(
    "--force",
    choices=["accept", "refuse"],
    default=None,
    help="Skips user confirmation when overwriting files and auto-accepts or auto-refuses all requests."
)


class ErrorMsg(errors.ErrorMsgBase):
    FILE_INCOMPLETE = "{} is incomplete."
    FILE_EMPTY = "{} is empty."
    FILE_MISSING = "{} does not exist."
    PERMISSION_ERROR = "Cannot access {}."
    STANDARD_MOODS_USED = "Standard mood set (rad, good, neutral, bad, awful) will be used."
    DECODE_ERROR = "Error while decoding {}"
    NOT_A_FILE = "{} is not a file."
    CSV_ALL_FIELDS_PRESENT = "All expected columns are present in the CSV file columns."
    CSV_FIELDS_MISSING = "The following expected columns are missing: {}"
    COUNT_ROWS = "{} rows of data found in {}. Of that {} were processed correctly."


class MissingValuesInRowError(utils.CustomException):
    """If a CSV row does not have enough values needed to create an entry."""


class CannotAccessFileError(utils.CustomException):
    """The file could not be accessed."""


class CannotAccessJournalError(CannotAccessFileError):
    """The journal CSV could not be accessed or parsed."""


class CannotAccessCustomMoodsError(CannotAccessFileError):
    """The custom moods JSON could not be accessed or parsed."""


class InvalidDataInFileError(utils.CustomException):
    """The file does not follow the expected structure."""


class NoDestinationSelectedError(utils.CustomException):
    """You have not specified where to output the files when instantiating this Librarian object."""


def create_and_open(filename: str, mode: str) -> IO:
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    return open(filename, mode, encoding="UTF-8")


# I've found a term that describes what this class does - it is a Director - even sounds similar to Librarian
# https://refactoring.guru/design-patterns/builder
class Librarian:
    """
    Orchestrates the entire process of parsing CSV & passing data to objects specialised to handle it as a journal.
    The chain of command looks like this:

    ``Librarian`` -> :class:`DatedEntriesGroup` -> :class:`DatedEntries`

    ---

    **How to process the CSV**

    User only needs to instantiate this object and pass the appropriate arguments.
    The processing does not require invoking any other functions. Functions of this class are therefore mostly private.

    ---

    **How to output the journal**

    TODO: add missing documentation
    """

    def __init__(self,
                 path_to_file: str,  # the only crucial parameter at this stage
                 path_to_output: str = None,
                 path_to_moods: str = None):
        """
        :param path_to_file: The path to the CSV file for processing.
        :param path_to_output: The path for outputting processed data as markdown files.
         If user does not provide the output path, no output functionality will work.
        :raises CannotAccessFileError: if any problems occur during accessing or decoding the CSV file.
        :param path_to_moods: The path for a custom mood set file.
        """
        self.__logger = logging.getLogger(self.__class__.__name__)
        self.__known_dates: dict[str, DatedEntriesGroup] = {}

        # Let's start processing the file
        # ---
        # 1. Parse the path_to_moods JSON for a custom mood set, if possible, or otherwise use standard mood set
        #
        # P.S Why am I starting first with moods? Because process_file first checks if it has moods installed.
        try:
            self.__mood_set = self.__create_mood_set(path_to_moods)
        except CannotAccessFileError as err:
            raise CannotAccessCustomMoodsError from err

        # 2. Access the CSV file and get all the rows with content
        #    then pass the data to specialised data objects that can handle them in a structured way
        # TODO: Deal with files that are valid but at the end of parsing have zero lines successfully parsed
        try:
            self.__process_file(path_to_file)
        except (CannotAccessFileError, InvalidDataInFileError) as err:
            raise CannotAccessJournalError from err

        # Ok, if no exceptions were raised so far, the file is good, let's go through the rest of the attributes
        self.__destination = path_to_output

    def __create_mood_set(self, json_file: str = None) -> 'Moodverse':
        """
        Overwrite the standard mood-set with a custom one. Mood-sets are used in colour-coding each dated entry.

        :param json_file: path to the .JSON file with a non-standard mood set.
         Should have five keys: ``rad``, ``good``, ``neutral``, ``bad`` and ``awful``.
         Each of those keys should hold an array of any number of strings indicating various moods.
         **Example**: ``[{"good": ["good"]},...]``
        :returns: reference to the :class:`Moodverse` object
        """
        if json_file:
            exp_path = utils.expand_path(json_file)
            try:
                with open(exp_path, encoding="UTF-8") as file:
                    custom_mood_set_from_file = json.load(file)
            except FileNotFoundError as err:
                msg = ErrorMsg.print(ErrorMsg.FILE_MISSING, exp_path)
                self.__logger.warning(msg)
                raise CannotAccessFileError(msg) from err
            except PermissionError as err:
                msg = ErrorMsg.print(ErrorMsg.PERMISSION_ERROR, exp_path)
                self.__logger.warning(msg)
                raise CannotAccessFileError(msg) from err
            except json.JSONDecodeError as err:
                msg = ErrorMsg.print(ErrorMsg.DECODE_ERROR, exp_path)
                self.__logger.warning(msg)
                raise CannotAccessFileError(msg) from err
        else:
            custom_mood_set_from_file = None

        # the command works with or without the argument
        # - Case 1: no argument = default mood-set
        # - Case 2: argument passed, but invalid = default mood-set
        # - Case 3: argument passed, it is valid = default mood-set expanded by the custom mood-set
        return Moodverse(custom_mood_set_from_file)

    # TODO: should return a tuple of { lines_processed_correctly, all_lines_processed }
    def __process_file(self, filepath: str) -> bool:
        """
        Validates CSV file and processes it into iterable rows.
        :param filepath: path to CSV to be read
        :raises CannotAccessFileError: if any problems occur during accessing the CSV file.
        :raises InvalidDataInFileError: if any problems occur during parsing the CSV file.
        :returns: True if parsed > 0, False otherwise
        """
        if not self.__mood_set.has_custom_moods:
            self.__logger.info(ErrorMsg.print(ErrorMsg.STANDARD_MOODS_USED))

        # Let's determine if the file can be opened
        # ---
        try:
            file = open(filepath, encoding='UTF-8')
        # File has not been found
        except FileNotFoundError as err:
            msg = ErrorMsg.print(ErrorMsg.FILE_MISSING, filepath)
            self.__logger.critical(msg)
            raise CannotAccessFileError(msg) from err
        # Insufficient permissions to access the file
        except PermissionError as err:
            msg = ErrorMsg.print(ErrorMsg.PERMISSION_ERROR, filepath)
            self.__logger.critical(msg)
            raise CannotAccessFileError(msg) from err
        # Other error that makes it impossible to access the file
        except OSError as err:
            self.__logger.critical(OSError)
            raise CannotAccessFileError from err

        # If the code reaches here, the program can access the file.
        # Now let's determine if the file's contents are actually usable
        # ---

        with file:
            # Is it a valid CSV?
            try:
                # strict parameter throws csv.Error if parsing fails
                raw_lines = csv.DictReader(file, delimiter=',', quotechar='"', strict=True)
            except csv.Error as err:
                msg = ErrorMsg.print(ErrorMsg.DECODE_ERROR, filepath)
                self.__logger.critical(msg)
                raise InvalidDataInFileError(msg) from err

            # Does it have all the fields? Push any missing field into an array for later reference
            # Even if only one column from the list below is missing in the CSV, it's a problem while parsing later
            expected_fields = [
                "full_date",
                "date",
                "weekday",
                "time",
                "mood",
                "activities",
                "note",
                "note_title",
                "note"
            ]

            # Let's have a look at what columns we have in the parsed CSV
            # It seems that even files with random bytes occasionally pass through previous checks with no errors
            # Therefore this 'try' block is also necessary, we do not know if the entire file is now fault-free
            try:
                missing_strings = [
                    expected_field for expected_field in expected_fields if expected_field not in raw_lines.fieldnames
                ]
            except (csv.Error, UnicodeDecodeError) as err:
                msg = ErrorMsg.print(ErrorMsg.DECODE_ERROR, filepath)
                self.__logger.critical(msg)
                raise InvalidDataInFileError(msg) from err

            if not missing_strings:
                self.__logger.debug(ErrorMsg.print(ErrorMsg.CSV_ALL_FIELDS_PRESENT))
            else:
                msg = ErrorMsg.print(
                    ErrorMsg.CSV_FIELDS_MISSING,
                    ', '.join(missing_strings)  # which ones are missing - e.g. "date, mood, note"
                )
                self.__logger.critical(msg)
                raise InvalidDataInFileError(msg)

            # # Does it have any rows besides the header?
            # # If the file is empty or only has column headers, exit immediately
            # try:
            #     next(raw_lines)
            # except StopIteration:
            #     msg = ErrorMsg.print(ErrorMsg.FILE_EMPTY, filepath)
            #     self.__logger.critical(msg)
            #     raise InvalidDataInFileError(msg)

            # If the code has reached this point and has not exited, it means both file and contents have to be ok
            # Processing
            # ---
            lines_parsed = 0
            lines_parsed_successfully = 0
            for line in raw_lines:
                line: dict[str]
                try:
                    lines_parsed += self.__process_line(line)
                except MissingValuesInRowError:
                    pass
                else:
                    lines_parsed_successfully += 1

            # Report back how many lines were parsed successfully out of all tried
            self.__logger.info(ErrorMsg.print(
                ErrorMsg.COUNT_ROWS, str(lines_parsed), filepath, str(lines_parsed_successfully))
            )

            # If at least one line has been parsed, the following return resolves to True
            return bool(lines_parsed)

    # TODO: I guess it is more pythonic to raise exceptions than return False if I cannot complete the task
    # TODO: this has to be tested
    # https://eli.thegreenplace.net/2008/08/21/robust-exception-handling/
    def __process_line(self, line: dict[str]) -> bool:
        """
        Goes row-by-row and passes the content to objects specialised in handling it from a journaling perspective.
        :raises MissingValuesInRowError: if the row in CSV lacks enough commas to create 8 cells. It signals a problem.
        :param line: a dictionary with values from the currently processed CSV line
        :return: True if all columns had values for this CSV ``line``, False otherwise
        """
        # Does each of the 8 columns have values for this row?
        if len(line) < 8:
            # Oops, not enough values on this row, the file might be corrupted?
            msg = ErrorMsg.print(ErrorMsg.FILE_INCOMPLETE, str(line))
            self.__logger.warning(msg)
            raise MissingValuesInRowError(msg)
        # Let DatedEntriesGroup handle the rest and increment the counter (True == 1)
        try:
            self.access_date(line["full_date"]).create_dated_entry_from_row(line)
        except (dated_entries_group.TriedCreatingDuplicateDatedEntryError,
                dated_entries_group.IncompleteDataRow,
                dated_entries_group.InvalidDateError,
                ValueError):
            return False
        return True

    def access_date(self, target_date: str) -> DatedEntriesGroup:
        """
        Accesses an already existing or creates a new :class:`DatedEntriesGroup` for the specified ``target_date``.
        :raises ValueError: if ``target_date`` is an invalid Date as indicated by :class:`Date` object
        :param target_date: the date for which a unique :class:`DatedEntriesGroup` object should be created or accessed.
        :return: reference to :class:`DatedEntriesGroup` object
        """
        try:
            date_lookup = dated_entries_group.Date(target_date)
        except dated_entries_group.InvalidDateError as err:
            raise ValueError from err

        if str(date_lookup) in self.__known_dates:
            return self.__known_dates[str(date_lookup)]
        else:
            new_obj = DatedEntriesGroup(str(date_lookup), self.__mood_set)
            self.__known_dates[str(date_lookup)] = new_obj
            return new_obj

    def output_all(self):
        """
        Loops through known dates and calls :class:`DatedEntriesGroup` to output its contents inside the destination.
        :raises NoDestinationSelectedError: when the parent object has been instantiated without a destination set.
        """
        if self.__destination is None:
            raise NoDestinationSelectedError

        for known_date in self.__known_dates.values():
            # "2022/11/09/2022-11-09.md"
            filename = str(known_date.date) + ".md"
            filepath = "/".join([self.__destination, known_date.date.year, known_date.date.month, filename])
            with create_and_open(filepath, 'a') as file:
                known_date.output(file)

    # Use a dunder overload of getitem to access groups in either way
    # 1. my_librarian["2022-10-10"]
    # 2. my_librarian.access_date("2022-10-10")
    def __getitem__(self, item: str) -> DatedEntriesGroup:
        ref = self.access_date(item)
        return ref

    @property
    def current_mood_set(self):
        return self.__mood_set



"""Parse a Daylio CSV into an Obsidian-compatible .MD file"""
import logging
import sys

from src.config import options
from src.librarian import Librarian

logger = logging.getLogger(__name__)

# Compile global settings
# ---
# Read arguments from console and update the global_settings accordingly
options.parse_console(sys.argv[1:])  # [1:] skips the program name, such as ["foo.py", ...]

# And now let's start processing
# ---
Librarian(path_to_file=options.filepath)



"""
dated_entry focuses on building the individual entries, made at a particular moment, as objects.
It is the most atomic level of the journaling process.

Here's a quick breakdown of what is the specialisation of this file in the journaling process:
all notes -> notes written on a particular date -> _A PARTICULAR NOTE_
"""
from __future__ import annotations

import logging
import re
import io
from typing import Match

import typing

from src import errors
from src import utils
from src.config import options
from src.entry.mood import Moodverse

# Adding DatedEntry-specific options in global_settings
dated_entry_settings = options.arg_console.add_argument_group(
    "Dated Entries",
    "Handles how entries should be formatted"
)
dated_entry_settings.add_argument(
    "--tags",
    help="Tags in the YAML of each note.",
    nargs='*',  # this allows, for example, "--tags one two three" --> ["one", "two", "three"]
    default="daily",
    dest="TAGS"
)
dated_entry_settings.add_argument(
    "--prefix",  # <here's your prefix>YYYY-MM-DD.md so remember about a separating char
    default='',
    help="Prepends a given string to each entry's header."
)
dated_entry_settings.add_argument(
    "--suffix",  # YYYY-MM-DD<here's your suffix>.md so remember about a separating char
    default='',
    help="Appends a given string at the end of each entry's header."
)
dated_entry_settings.add_argument(
    "--tag_activities", "-a",
    action="store_true",  # default=True
    help="Tries to convert activities into valid tags.",
    dest="ACTIVITIES_AS_TAGS"
)
dated_entry_settings.add_argument(
    "-colour", "--color",
    action="store_true",  # default=True
    help="Prepends a colour emoji to each entry depending on mood.",
    dest="colour"
)
dated_entry_settings.add_argument(
    "--header",
    type=int,
    default=2,
    help="Headings level for each entry.",
    dest="HEADER_LEVEL"
)
dated_entry_settings.add_argument(
    "--csv-delimiter",
    default="|",
    help="Set delimiter for activities in Daylio .CSV, e.g: football | chess"
)


class IsNotTimeError(utils.CustomException):
    msg = "Expected HH:MM (+ optionally AM/PM suffix) but got {} instead."

    def __init__(self, tried_time: str):
        super().__init__(type(self).msg.format(tried_time))


def is_time_format_valid(string: str) -> Match[str] | None:
    """
    Is the time format of :param:`str` valid?
    :param string: time to check
    :return: ``True`` if :param:`str` follows the ``HH:MM`` format, with optional AM/PM appended, ``False`` otherwise
    """
    return re.compile(r'^([0-1]?[0-9]|2[0-3]):[0-5][0-9]($|\sAM|\sPM)').match(string)


def is_time_range_valid(string: str) -> bool:
    """
    Is the time range of :param:`str` valid?
    :param string: time to check
    :return: ``True`` if hour and minute ranges are both ok, ``False`` otherwise
    """
    time_array = string.strip().split(':')

    # Check if it's in 12-hour format (AM/PM) or 24-hour format
    if 'AM' in string or 'PM' in string:
        is_hour_ok = 0 <= int(time_array[0]) <= 12
    else:
        is_hour_ok = 0 <= int(time_array[0]) < 24

    # Minutes can be checked irrespective of AM/PM/_ notation
    is_minutes_ok = 0 <= int(time_array[1][:2]) < 60

    return all((is_hour_ok, is_minutes_ok))


class ErrorMsg(errors.ErrorMsgBase):
    INVALID_MOOD = "Mood {} is missing from a list of known moods. Not critical, but colouring won't work on the entry."
    WRONG_TIME = "Received {}, expected valid time. Cannot create this entry without a valid time."
    WRONG_ACTIVITIES = "Expected a non-empty list of activities. In that case just omit this argument in function call."
    WRONG_TITLE = "Expected a non-empty title. Omit this argument in function call rather than pass a falsy string."
    WRONG_NOTE = "Expected a non-empty note. Omit this argument in function call rather than pass a falsy string."


class Time:
    """
    Hour and minutes of a particular moment in time. Validates the time string on instantiation.
    str(instance) returns the valid time in the ``HH:MM`` format.
    :raises IsNotTimeError: if string is not a valid time in ``HH:MM`` format (either AM/PM or 24h)
    """

    def __init__(self, string: str):
        """
        Upon instantiation checks if the time is valid.
        Used in :class:`DatedEntry` to create an instance of this class.
        :raises IsNotTime: if string is not a valid time in ``HH:MM`` format with optional AM/PM appended
        :param string: time in ``HH:MM`` format - can have AM/PM appended
        """
        self.__logger = logging.getLogger(self.__class__.__name__)

        # OK
        if is_time_format_valid(string.strip()) and is_time_range_valid(string.strip()):
            time_array = string.strip().split(':')
            self.__hour = time_array[0]
            self.__minutes = time_array[1]

        # NOT OK
        else:
            self.__logger.warning(IsNotTimeError.msg.format(string))
            raise IsNotTimeError(string)

    def __str__(self) -> str:
        """
        :return: Outputs its hour and minutes attributes as a string in valid time format - HH:MM.
        """
        return ':'.join([self.__hour, self.__minutes])


class DatedEntry(utils.Core):
    """
    Journal entry.
    **A journal entry cannot exists without:**

    * Time it was written at, as :class:`Time`
    * Mood, that is -  a dominant emotional state during that particular moment in time.

    **Other, optional attributes:**

    * title
    * note
    * activities performed during or around this time

    :raises ValueError: if at least one of the required attributes cannot be set properly
    """

    def __init__(self,
                 time: str,
                 mood: str,
                 activities: str = None,
                 title: str = None,
                 note: str = None,
                 override_mood_set: Moodverse = Moodverse()):
        """
        :param time: Time at which this note was created
        :param mood: Mood felt during writing this note
        :param activities: (opt.) Activities carried out around or at the time of writing the note
        :param title: (opt.) Title of the note
        :param note: (opt.) The contents of the journal note itself
        :param override_mood_set: Set if you want to use custom :class:`Moodverse` for mood handling
        """
        # TODO: have to test the whole instantiation function again after refactoring
        self.__logger = logging.getLogger(self.__class__.__name__)

        # Processing required properties
        # ---
        # TIME
        # ---
        try:
            super().__init__(Time(time))
        except IsNotTimeError as err:
            errors.ErrorMsgBase.print(ErrorMsg.WRONG_TIME, time)
            raise ValueError from err

        # ---
        # MOOD
        # ---
        if not mood:
            raise ValueError
        # Check if the mood is valid - i.e. it does exist in the currently used Moodverse
        if not override_mood_set.get_mood(mood):
            errors.ErrorMsgBase.print(ErrorMsg.INVALID_MOOD, mood)
        # Warning is enough, it just disables colouring so not big of a deal
        self.__mood = mood

        # Processing other, optional properties
        # ---
        # Process activities
        # ---
        self.__activities = []
        if activities:
            working_array = utils.strip_and_get_truthy(activities, options.csv_delimiter)
            if len(working_array) > 0:
                for activity in working_array:
                    self.__activities.append(utils.slugify(
                        activity,
                        options.tag_activities
                    ))
            else:
                errors.ErrorMsgBase.print(ErrorMsg.WRONG_ACTIVITIES)
        # ---
        # Process title
        # ---
        self.__title = utils.slice_quotes(title) if title else None
        if not title:
            errors.ErrorMsgBase.print(ErrorMsg.WRONG_TITLE)
        # ---
        # Process note
        # ---
        self.__note = utils.slice_quotes(note) if note else None
        if not note:
            errors.ErrorMsgBase.print(ErrorMsg.WRONG_NOTE)

    def output(self, stream: io.IOBase | typing.IO) -> int:
        """
        Write entry contents directly into the provided buffer stream.
        It is the responsibility of the caller to handle the stream afterward.
        :raises utils.StreamError: if the passed stream does not support writing to it.
        :raises OSError: likely due to lack of space in memory or filesystem, depending on the stream
        :param stream: Since it expects the base :class:`io.IOBase` class, it accepts both file and file-like streams.
        :returns: how many characters were successfully written into the stream.
        """
        if not stream.writable():
            raise utils.StreamError

        chars_written = 0
        # HEADER OF THE NOTE
        # e.g. "## great | 11:00 AM | Oh my, what a night!"
        # options.header is an int that multiplies the # to create headers in markdown
        header_elements = [
            options.header * "#" + ' ' + self.__mood,
            self.time,
            self.__title
        ]
        header = ' | '.join([el for el in header_elements if el is not None])
        chars_written += stream.write(header)
        # ACTIVITIES
        # e.g. "bicycle skating pool swimming"
        if len(self.__activities) > 0:
            chars_written += stream.write("\n" + ' '.join(self.__activities))
        # NOTE
        # e.g. "Went swimming this evening."
        if self.__note is not None:
            chars_written += stream.write("\n" + self.__note)

        return chars_written

    @property
    def mood(self):
        return self.__mood

    @property
    def activities(self):
        return self.__activities

    @property
    def title(self):
        return self.__title

    @property
    def note(self):
        return self.__note

    @property
    def time(self):
        return str(self.uid)



"""
Contains universally useful functions
"""
import logging
import os
import re
from typing import Any, List

from src import errors


class ErrorMsg(errors.ErrorMsgBase):
    INVALID_OBSIDIAN_TAGS = "You want your activities as tags, but {} is invalid."


class Core:
    def __init__(self, uid):
        self.__uid = uid

    def __bool__(self):
        return self.__uid is not None

    def __str__(self):
        return str(self.__uid)

    def __hash__(self):
        return hash(self.uid)

    @property
    def uid(self):
        return self.__uid


class CustomException(Exception):
    def __init__(self, message=None):
        super().__init__(message)
        self.message = message


class StreamError(CustomException):
    pass


def slugify(text: str, taggify: bool) -> str:
    # noinspection SpellCheckingInspection
    """
    Simple slugification function to transform text. Works on non-latin characters too.
    """
    logger = logging.getLogger(__name__)
    text = str(text).lower().strip()  # get rid of trailing spaces left after splitting activities apart from one string
    text = re.sub(re.compile(r"\s+"), '-', text)  # Replace spaces with -
    text = re.sub(re.compile(r"[^\w\-]+"), '', text)  # Remove all non-word chars
    text = re.sub(re.compile(r"--+"), '-', text)  # Replace multiple - with single -
    text = re.sub(re.compile(r"^-+"), '', text)  # Trim - from start of text
    text = re.sub(re.compile(r"-+$"), '', text)  # Trim - from end of text
    # Checks if the tag is actually a valid tag in Obsidian - still appends the hash even if not, but warns at least
    if taggify:
        if re.match('[0-9]', text):
            logger.warning(ErrorMsg.print(ErrorMsg.INVALID_OBSIDIAN_TAGS, text))
    return '#' + text if taggify else text


def expand_path(path: str) -> str:
    """
    Expand all %variables%, ~/home-directories and relative parts in the path. Return the expanded path.
    It does not use os.path.abspath() because it treats current script directory as root.
    """
    # Gets full path, resolving things like ../
    return os.path.realpath(
        # Expands the tilde (~) character to the user's home directory
        os.path.expanduser(
            os.path.expandvars(path)
        )
    )


def slice_quotes(string: str) -> str | None:
    """
    Gets rid of initial and terminating quotation marks inserted by Daylio
    :param string: string to be sliced
    :returns: string without quotation marks in the beginning and end of the initial string, or nothing if "" provided.
    """
    # only 2 characters? Then it is an empty cell, because Daylio wraps its values inside "" like so: "","","",""...
    return string.strip("\"").strip() if string and len(string) > 2 else None


def strip_and_get_truthy(delimited_string: str, delimiter: str) -> List[str]:
    """
    Pipe delimited strings may result in arrays that contain zero-length strings.
    While such strings in itself are falsy, any array that has them is automatically truthy, unfortunately.
    Therefore, I use list comprehension to discard such falsy values from an array and return the sanitised array.
    :returns: array without falsy values, even if it results in empty (falsy) array
    """
    # I need to separate returning into the guard statement and actual return because slice_quotes can produce null vals
    if delimited_string is None:
        return []

    sliced_del_string = slice_quotes(delimited_string)

    return [el for el in sliced_del_string.split(delimiter) if el] if sliced_del_string else []



import logging
import sys
from typing import Optional


class ColorHandler(logging.StreamHandler):
    # https://en.wikipedia.org/wiki/ANSI_escape_code#Colors
    GRAY8 = "38;5;8"
    GRAY7 = "38;5;7"
    ORANGE = "33"
    RED = "31"
    WHITE = "0"

    def emit(self, record):
        # We don't use white for any logging, to help distinguish from user print statements
        level_color_map = {
            logging.DEBUG: self.GRAY8,
            logging.INFO: self.GRAY7,
            logging.WARNING: self.ORANGE,
            logging.ERROR: self.RED,
            logging.CRITICAL: f"1;{self.RED}",  # Bold for critical errors
        }

        csi = f"{chr(27)}["  # control sequence introducer
        color = level_color_map.get(record.levelno, self.WHITE)

        # Apply the formatter to format the log message
        formatted_msg = self.format(record)

        self.stream.write(f"{csi}{color}m{formatted_msg}{csi}m\n")


# Create a console handler for the root logger
# noinspection SpellCheckingInspection
console_log_handler = ColorHandler(sys.stdout)
console_log_handler.setLevel(logging.DEBUG)

# noinspection SpellCheckingInspection
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)")
console_log_handler.setFormatter(formatter)

# Add the handlers to the root logger
logging.getLogger().addHandler(console_log_handler)


class ErrorMsgBase:
    """
    Used for common errors that will be logged by almost (if not all) loggers.
    Therefore, it is the base class of other Error child classes in specific files.
    It also provides a shorthand method for inserting variables into error messages - print().
    """
    # some common errors have been raised in scope into base class instead of child classes
    OBJECT_FOUND = "{}-class object found."
    OBJECT_NOT_FOUND = "{}-class object not found."
    FAULTY_OBJECT = "Called a {}-class object method but the object has been incorrectly instantiated."
    WRONG_VALUE = "Received {}, expected {}."

    @staticmethod
    def print(message: str, *args: str) -> Optional[str]:
        """
        Insert the args into an error message. If the error message expects n variables, provide n arguments.
        Returns a string with the already filled out message.
        """
        expected_args = message.count("{}")

        if len(args) != expected_args:
            logging.getLogger(__name__).warning(
                f"Expected {expected_args} arguments for \"{message}\", but got {len(args)} instead."
            )
            return None
        return message.format(*args)



"""
This file specialises in building the middleware between actual journal entries and the whole journal.
It creates and organises only those entries written on a particular date. This way they can be handled easier.

Here's a quick breakdown of what is the specialisation of this file in the journaling process:
all notes -> _NOTES WRITTEN ON A PARTICULAR DATE_ -> a particular note
"""
from __future__ import annotations

import io
import logging
import re

import typing

from src import dated_entry
from src import errors
from src import utils
from src.dated_entry import DatedEntry
from src.entry.mood import Moodverse
from src.config import options


class DatedEntryMissingError(utils.CustomException):
    """The :class:`DatedEntry` does not exist."""


class IncompleteDataRow(utils.CustomException):
    """Passed a row of data from CSV file that does not have all required fields."""


class InvalidDateError(utils.CustomException):
    """String is not a valid date. Check :class:`Date` for details."""


class TriedCreatingDuplicateDatedEntryError(utils.CustomException):
    """Tried to create object of :class:`DatedEntry` that would be a duplicate of one that already exists."""


class ErrorMsg(errors.ErrorMsgBase):
    CSV_ROW_INCOMPLETE_DATA = "Passed .csv contains rows with invalid data. Tried to parse {} as date."


class Date:
    """
    Day, month and year of a particular date. Validates the date string on instantiation.
    """
    _instances = {}  # Class variable to store instances based on date

    def __new__(cls, string: str):
        # Check if an instance for the given date already exists
        if string in cls._instances:
            return cls._instances[string]
        # If not, create a new instance
        instance = super(Date, cls).__new__(cls)
        cls._instances[string] = instance
        return instance

    def __init__(self, string: str):
        """
        :raises InvalidDateError: if :param:`string` is not a valid date (for example the month number > 12)
        :param string: on which entries have been created (`YYYY-MM-DD`)
        """
        # self.__logger = logging.getLogger(self.__class__.__name__)

        # does it have a valid format YYYY-MM-DD
        valid_date_pattern = re.compile(r'^\d{4}-\d{1,2}-\d{1,2}$')
        if not re.match(valid_date_pattern, string):
            raise InvalidDateError

        # does it have valid ranges? (year ranges are debatable)
        date_array = string.strip().split('-')
        if not all((
                1900 < int(date_array[0]) < 2100,
                0 < int(date_array[1]) < 13,
                0 < int(date_array[2][:2]) < 32)):
            raise InvalidDateError

        self.__year = date_array[0]
        self.__month = date_array[1]
        self.__day = date_array[2]

    def __str__(self) -> str:
        """
        :return: returns the valid date in the YYYY-MM-DD format. This is the superior format, end of discussion.
        """
        return '-'.join([self.__year, self.__month, self.__day])

    def __eq__(self, other: 'Date') -> bool:
        """Used only for comparing two :class:`Date` objects - itself and another one."""
        if isinstance(other, Date):
            return all((other.year == self.year,
                        other.month == self.month,
                        other.day == self.day))
        return super().__eq__(other)

    @property
    def year(self):
        return self.__year

    @property
    def month(self):
        return self.__month

    @property
    def day(self):
        return self.__day


class DatedEntriesGroup(utils.Core):
    """
    A particular date which groups entries written that day.

    Imagine it as a scribe, holding a stack of papers in his hand.
    The master Librarian knows each one of the scribes, including this one.
    However, the scribe knows only his papers. The papers contain all entries written that particular date.

    Truthy if it knows at least one :class:`DatedEntry` made on this :class:`Date`.
    """
    _instances = {}

    def __new__(cls, date: str, current_mood_set: Moodverse = Moodverse()):
        # Check if an instance for the given date already exists
        if date in cls._instances:
            return cls._instances[date]
        # If not, create a new instance
        instance = super(DatedEntriesGroup, cls).__new__(cls)
        cls._instances[date] = instance
        return instance

    def __init__(self, date, current_mood_set: Moodverse = Moodverse()):
        """
        :raises InvalidDateError: if the date string is deemed invalid by :class:`Date`
        :param date: The date for all child entries within.
        :param current_mood_set: Use custom :class:`Moodverse` or default if not provided.
        """
        self.__logger = logging.getLogger(self.__class__.__name__)

        # Try parsing the date and assigning it as your identification (uid)
        try:
            super().__init__(Date(date))
        # Date is no good?
        except InvalidDateError as err:
            msg = ErrorMsg.print(ErrorMsg.WRONG_VALUE, date, "YYYY-MM-DD")
            self.__logger.warning(msg)
            raise InvalidDateError(msg) from err

        # All good - initialise
        self.__known_entries_for_this_date: dict[str, DatedEntry] = {}
        self.__known_moods: Moodverse = current_mood_set

    def append_to_known(self, entry: DatedEntry) -> None:
        self.__known_entries_for_this_date[str(entry.uid)] = entry

    def create_dated_entry_from_row(self,
                                    line: dict[str, str]) -> dated_entry.DatedEntry:
        """
        :func:`access_dated_entry` of :class:`DatedEntry` object with the specified parameters.
        :raises TriedCreatingDuplicateDatedEntryError: if it would result in making a duplicate :class:`DatedEntry`
        :raises IncompleteDataRow: if ``line`` does not have ``time mood`` keys at the very least, or either is empty
        :raises ValueError: re-raises ValueError from :class:`DatedEntry`
        :param line: a dictionary of strings. Required keys: mood, activities, note_title & note.
        """
        # TODO: test case this
        # Try accessing the minimum required keys
        for key in ["time", "mood"]:
            try:
                line[key]
            except KeyError as err:
                raise IncompleteDataRow(key) from err
            # is it empty then, maybe?
            if not line[key]:
                raise IncompleteDataRow(key)

        # Check if there's already an object with this time
        # TODO: Daylio actually allows creating multiple entries and mark them as written at the same time
        if line["time"] in self.__known_entries_for_this_date:
            raise TriedCreatingDuplicateDatedEntryError

        # Instantiate the entry
        try:
            this_entry = dated_entry.DatedEntry(
                line["time"],
                line["mood"],
                activities=line["activities"],
                title=line["note_title"],
                note=line["note"],
                override_mood_set=self.__known_moods
            )
        except ValueError as err:
            raise ValueError from err

        self.append_to_known(this_entry)
        return this_entry

    def access_dated_entry(self, time: str) -> DatedEntry:
        """
        Retrieve an already existing DatedEntry object.
        :param time: any string, but if it's not a valid HH:MM format then I guarantee it won't be found either way
        :raises DatedEntryMissingError: if the entry specified in ``time`` does not exist
        :returns: :class:`DatedEntry`
        """
        try:
            ref = self.__known_entries_for_this_date[time]
        except KeyError as err:
            msg = ErrorMsg.print(ErrorMsg.OBJECT_NOT_FOUND, time)
            self.__logger.warning(msg)
            raise DatedEntryMissingError(msg) from err
        self.__logger.debug(ErrorMsg.print(ErrorMsg.OBJECT_FOUND, time))
        return ref

    def output(self, stream: io.IOBase | typing.IO) -> int:
        """
        Write entry contents of all :class:`DatedEntry` known directly into the provided buffer stream.
        It is the responsibility of the caller to handle the stream afterward.
        :raises utils.StreamError: if the passed stream does not support writing to it.
        :raises OSError: likely due to lack of space in memory or filesystem, depending on the stream
        :param stream: Since it expects the base :class:`io.IOBase` class, it accepts both file and file-like streams.
        :returns: how many characters were successfully written into the stream.
        """
        if not stream.writable():
            raise utils.StreamError

        chars_written = 0
        # THE BEGINNING OF THE FILE
        # when appending file tags at the beginning of the file, discard any duplicates or falsy strings
        # sorted() is used to have a deterministic order, set() was random, so I couldn't properly test the output
        valid_tags = sorted(set(val for val in options.tags if val))
        if valid_tags:
            # why '\n' instead of os.linesep?
            # > Do not use os.linesep as a line terminator when writing files opened in text mode (the default);
            # > use a single '\n' instead, on all platforms.
            # https://docs.python.org/3.10/library/os.html#os.linesep
            chars_written += stream.write("---" + "\n")
            chars_written += stream.write("tags: " + ",".join(valid_tags) + "\n")
            chars_written += stream.write("---" + "\n"*2)

        # THE ACTUAL ENTRY CONTENTS
        # Each DatedEntry object now appends its contents into the stream
        for entry in self.__known_entries_for_this_date.values():
            # write returns the number of characters successfully written
            # https://docs.python.org/3/library/io.html#io.TextIOBase.write
            if entry.output(stream) > 0:
                chars_written += stream.write("\n"*2)

        return chars_written

    @property
    def known_entries_from_this_day(self):
        return self.__known_entries_for_this_date

    @property
    def date(self):
        """
        :return: String in the format of YYYY-MM-DD that identifies this specific object of :class:`DatedEntryGroup`.
        """
        return self.uid



"""
Sets up all necessary options and arguments.
.
├── Librarian/
│   ├── processing/
│   │   └── filepath
│   └── outputting/
│       ├── destination
│       └── force
|
├── DatedGroup/
│   ├── processing
│   └── outputting
|
└── DatedEntries/
    ├── processing/
    │   └── csv_delimiter
    └── outputting/
        ├── prefix
        ├── suffix
        ├── colour
        └── header
"""
import argparse
import logging
from typing import List, Any

# Logging for config library
logger = logging.getLogger(__name__)


class SettingsManager:
    def __init__(self):
        # noinspection SpellCheckingInspection
        self.__console_arguments = argparse.ArgumentParser(
            fromfile_prefix_chars="@",
            prog="Daylio to Obsidian Parser",
            description="Converts Daylio .CSV backup file into Markdown files for Obsidian.",
            epilog="Created by DeutscheGabanna"
        )
        # regardless of whether you stick with these defaults or parse the console args, you get the same properties
        # useful for testing purposes when you do not want to invoke argparse
        # ---
        # Librarian
        # ---
        self.filepath = None
        self.destination = None
        # TODO: Force-argument does nothing yet.
        self.force = None
        # ---
        # Dated Entry
        # ---
        self.csv_delimiter = '|'
        self.header = 2
        self.tags = ["daily"]
        self.prefix = ''
        self.suffix = ''
        self.tag_activities = True
        self.colour = True
        # TODO: User should be able to set verbosity level in logging

    @property
    def arg_console(self) -> argparse.ArgumentParser:
        """
        Retrieves the :class:`argparse.ArgumentParser` object from :class:`SettingsManager` so you can modify it.
        :return: :class:`argparse.ArgumentParser`
        """
        return self.__console_arguments

    def parse_console(self, args: List[Any]) -> None:
        """
        Configures SettingsManager by accessing the console and retrieving the arguments used to run the script.
        :param args: either console arguments from sys.argv or spoofed ones
        """
        # namespace=self adds the properties to the SettingsManager obj, instead of creating a new Namespace obj
        # Without namespace=self
        # ---
        # - SettingsManager
        #   - Namespace obj that holds actual settings
        #       - foo = foo
        #       - bar = bar
        # With namespace=self
        # - SettingsManager
        #   - foo = foo
        #   - bar = bar
        self.__console_arguments.parse_args(args=args, namespace=self)


# Global configuration
# ---
options = SettingsManager()

# Add some common options
options.arg_console.add_argument(
    '--version',
    action='version',
    version='%(prog)s 3.0'
)
